/*
Module : CMapi.cpp
Purpose: Defines the implementation for an C++ wrapper class for sending an email using simple MAPI
Created: PJN / 11-05-1999
History: PJN / 05-12-1999 1. Fixed potential problem where CMapi is used in a console app which was giving an ASSERT
                          2. Fixed an assert which can occur if you dismiss the login dialog when doing an 
                          interactive MAPI logon
         PJN / 28-03-2000 1. Provision of a sample app which implements a mini mail sender.
                          2. Added a CMapiMessage::AddMultipleRecipients method similar to my SMTP class.
         PJN / 02-04-2000 1. Password field in the configuration dialog now uses ES_PASSWORD edit control style.
                          2. Password value used for Mapi logon is now stored in the registry encrypted instead of as plain text.
         PJN / 21-04-2000 1. Added an optional parameter to CMapi::SendMessage to allow you to decide whether or not addresses 
                          should be resolved.
         PJN / 22-06-2000 1. Fixed a bug in CMapiMessage::AddMultipleRecipients where BCC recipients were not being added 
                          correctly.
         PJN / 18-07-2000 1. Fixed check for MAPI being installed when client apps create an incorrect "MAPI" entry in win.ini
                          or the registry, Thanks to Chris Raethke for spotting this.
                          2. A "SharedLogon" method has been provided which tries to acquire a shared mapi session
                          3. SendMessage method has been extended to allow the message to be interactively edited.
         PJN / 25-01-2003 1. Updated copyright messages.
                          2. Made CMapiSession destructor virtual
                          3. Made a new CMapiRecipient class. This allows finer control over setting the display name
                          and the address individually. Thanks to the people on the discussion board on codeproject.com
                          for reporting this problem. The problem arose when Outlook Express displays its security dialog 
                          about an unattended mapi mail being sent. In this dialog the "To" fields were all empty. It seems
                          that OE only displays the addresses in this dialog. Perhaps because so many viruses abound the 
                          address is more relevant that the actual friendly name.
         PJN / 05-07-2003 1. Fixed a bug in the ordering of the string parameters to the CMapiRecipient constructor. Thanks
                          to "Louchkov" for spotting this problem.
         PJN / 22-12-2003 1. Added support for reading mail thro MAPI. Thanks to Marcel Scherpenisse for this addition.
         PJN / 05-02-2004 1. Added a pragma message to inform uses about including afxtempl.h in the precompiled header so as
                          to avoid compiler errors.
         PJN / 05-05-2004 1. Added support for deleting emails via MAPI. Thanks to Marcel Scherpenisse for this addition.
                          2. Also added the helper function CMapiMessage::AddMultipleAttachments. Again thanks to Marcel 
                          Scherpenisse for this addition.
         PJN / 05-06-2004 1. Added a parameter to the Logon methods to allow emails to be download (i.e. MAPI_FORCE_DOWNLOAD).
                          Thanks to "caowen" for this update.
         PJN / 30-04-2005 1. Fix for a bug where the MapiMessage::flFlags value was not being set when a message was being
                          sent. Failure to set the flags value meant that CMapi did not support the "MAPI_RECEIPT_REQUESTED"
                          flag. Thanks to Marcel Scherpenisse for reporting this bug.
         PJN / 13-08-2005 1. Fixed a number of bugs related to the use of MAPI_FORCE_DOWNLOAD setting. Thanks to Albert van Peppen
                          for this.
                          2. Optimized AddMultipleRecipients and AddMultipleAttachments function somewhat by using new 
                          CString::Trim method when available as well as avoiding addresses which are empty. Again thanks 
                          to Albert van Peppen for this.
                          3. MAPI_FORCE_DOWNLOAD is now set if Logon method has to attempt a shared logon. Again thanks to
                          Albert van Peppen for this.
                          4. Fixed a bug in SharedLogon where the MAPI_FORCE_DOWNLOAD flag was not being passed. Again thanks to
                          Albert van Peppen for this.
                          5. Replaced calls to ZeroMemory with memset.
                          6. Class now supports sending HTML emails (well kindof!). To achieve this a new parameter "bSendAsHTML" is 
                          included in the Send method. Please note that the HTML is sent as an additional attachment before all the 
                          other attachments and its disposition type is still marked as an attachment (at least using the latest 
                          version of Outlook on Windows XP). In addition some mail clients (such as Outlook Express) will still interpret 
                          this as an attachment as opposed to inline. If you are not happy with this then you will need to look into 
                          using Extended MAPI which this class does not provide for (and will never) or better yet, you should use SMTP 
                          directly, since HTML email is really part of the SMTP suite of protocols. For example you could use the 
                          author's own CPJNSMTPConnection class. Again thanks to to Albert van Peppen for this.   
                          7. Sample app now allows multiple attachments to be specified as edit box is now not read only.
                          8. Fixed a duplicate mnemonics issue in the main dialog in the sample app
                          9. File dialog which appears in the sample app when you add an attachment not includes a standard file name
                          filter.
                          10. Fixed a number of compiler warnings when the code is compiled using the new Force conformance in for loop
                          setting in Visual Studio 2003.
                          11. Fixed a number of compiler warnings when the code is compiled using the Detect 64-bit portability issues
                          in Visual Studio 2003. Please note that this means the codes requires a recentish copy of the Platform SDK
                          to be installed if you are compiling the code on Visual C++ 6.0
         PJN / 14-08-2005 1. Fixed problems when the code is compiled for UNICODE. The issue was that the Send method was using the 
                          various T2A macros which allocates a stack based string. This was causing issues because the recipients and 
                          attachments array for the MAPI message was being allocated in loops which was using scoping. This scoping was 
                          causing the stack based strings to be deallocated before the "MAPISendMail" method was called. This issue has 
                          now been fixed by the introduction of new generic heap based macros in PJNConv.h which allocates the converted 
                          string using the C++ new operator. To ensure orderly cleanup of this heap memory, the MapiMessage structure is 
                          now allocated using a hidden C++ class "CMapiMessageCleanup" which cleans up this memory automatically in its 
                          destructor. These macros are quite generic and can be used independent of the CMapi class in places where your 
                          code has the handle this exact same issue. 
         PJN / 19-08-2005 1. Fixed a bug in the calculation of the flags parameter in the call to acquire a shared logon in the Logon 
                          method. Thanks to Albert van Peppen for reporting this problem.
         PJN / 23-07-2006 1. Updated copyright details.
                          2. Optimized CMapiMessage constructor code.
                          3. Optimized CMapiSession constructor code.
                          4. Code now uses newer C++ style casts instead of C style casts.
                          5. Updated documentation to use the same style as the web site.
                          6. Fixed a bug freeing up strings in the sample app's CEnumMessagesDlg::OnDestroy method
                          7. Updated the code to clean compile on VC 2005
                          8. Sample app now allows you to specify no To address if the interactive setting is specified
                          9. Removed an unnecessary assert from CMapiMessage::AddMultipleRecipients and 
                          CMapiMessage::AddMultipleAttachments.
                          10. Updated some assert logic in CMapiSession::Send to not assert on the recipient count if we are sending 
                          a message "Interactive". Thanks to Sergiu Scobici for reporting this issue.
                          11. Refactored some code in CMapiSession::Send into CMapiSession::SetupRecipients. Again thanks to Sergiu 
                          Scobici for reporting this 
         PJN / 26-07-2006 1. Minor update to Logoff to avoid 2 asserts which occur when MAPI is not installed on client machines. 
                          Thanks to Sergiu Scobici for reporting this
         PJN / 22-12-2006 1. Further updates to fix compile issues for the sample app on VC 2005
                          2. Subject and body and email messages are now allocated using the H_T2A macro. This will avoid problems
                          with large CString values for these fields
                          3. Fixed a bug in CMapiSession::Logoff where the return value would always be TRUE
                          4. A empty profile name is now allowed in the sample app.
                          5. Configuration dialog is now not shown unnecessarily when you send an email in the sample app.
                          6. Logon is now only attempted in the sample app if the Interactive flag is not set
                          7. Addition of a CMAPI_EXT_CLASS preprocessor macro to allow the classes to be more easily added to an 
                          extension dll
         PJN / 01-01-2007 1. Updated copyright details.
                          2. Fixed incorrect usage of preprocessor macro in CMapiMessage::AddMultipleRecipients method
         PJN / 15-09-2008 1. Updated copyright details.
                          2. Code now compiles cleanly using Code Analysis (/analyze)
                          3. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          4. Updated sample app to clean compile on VC 2008
                          5. The code has now been updated to support VC 2005 or later only. 
                          6. Removed VC 6 style AppWizard comments from the code.
                          7. Renamed CMapiMessage::AddMultipleAttachments to CMapiMessage::ParseMultipleRecipients and reworked its 
                          parameters to be in line with the author's SMTP class.
                          8. Reworked the way the code handles allocations for the ASCII strings required by simple MAPI. Now these
                          are allocated in the message itself rather than using raw heap allocations via the old H_T2A macro
                          9. AddMultipleRecipients has been renamed to ParseMultipleRecipients in line with the author's SMTP class
                          10. The CMapiRecipient constructor which takes one string parameter now supports breaking the address
                          into email address and friendly name parts
                          11. AddMultipleAttachments now returns the number of attachments added
         PJN / 06-11-2009 1. Updated the zip file to remove the now defunct PJNConv.h header file
                          2. The sample app now uses DPAPI to encrypt the profile & password configuration settings
         PJN / 01-04-2011 1. Updated copyright details.
                          2. Replaced use of CT2A with CStringA
                          3. Updated sample app to clean compile on VC 2010
         PJN / 04-10-2012 1. Updated copyright details
                          2. Updated code to clean compile on VC 2012
                          3. Code no longer uses SearchPath or LoadLibrary without an absolute path when loading MAPI DLL. This
                          avoids DLL planting security issues.
         PJN / 21-03-2015 1. Updated code to clean compile on VC 2013
                          2. Updated copyright details.
                          3. All CWnd* parameters have been replaced with HWND parameters.
                          4. The classes have now been put in a CMapi namespace and have been renamed to CRecipient, CMessage & 
                          CSession
                          5. Reworked the classes to optionally compile without MFC. By default the classes now use STL classes 
                          and idioms but if you define CMAPI_MFC_EXTENSIONS the classes will revert back to the MFC behaviour.
                          6. Reworked all the classes so that all the exposed strings are now ASCII. The onus is now on client
                          code to convert the data to ASCII when integrating with the CMapi classes.
         PJN / 23-11-2015 1. Updated code to clean compile on VC 2015
         PJN / 11-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will 
                          now only compile on VC 2017 or later.
                          3. Removed all code which supported CMAPI_MFC_EXTENSIONS code path
                          4. Replaced BOOL with bool throughout the code
         PJN / 14-10-2018 1. Reimplemented the CMapi::CMessage::AddMultipleAttachments method.
                          2. Renamed the the CMapi::CMessage::ParseMultipleRecipients method to AddMultipleRecipients. In addition
                          this method no longer clears the recipients vector at the start of the method.
         PJN / 22-04-2019 1. Updated copyright details
                          2. Updated the code to clean compile on VC 2019
         PJN / 11-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 29-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 19-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 04-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
                          2. Added support for MapiSendMailW. This allows the code to support Unicode formatted
                          emails.
         PJN / 19-02-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations
         PJN / 11-05-2023 1. Updated copyright details
                          2. Updated modules to indicate that it needs to be compiled using /std:c++17. Thanks to Martin Richter
                          for reporting this issue.
         PJN / 23-09-2023 1. Updated SAL across the code.

Copyright (c) 1999 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


////////////////////////// Includes ///////////////////////////////////////////

#include "stdafx.h"
#include "CMapi.h"
#ifndef __ATLBASE_H__
#pragma message("To avoid this message, please put atlbase.h in your pre compiled header (normally stdafx.h)")
#include <atlbase.h>
#endif
#ifndef __ATLFILE_H__
#pragma message("To avoid this message, please put atlfile.h in your pre compiled header (normally stdafx.h)")
#include <atlfile.h>
#endif


////////////////////////// Implementation /////////////////////////////////////

CMapi::CRecipient::CRecipient(_In_ String sAddress)
{
  //The local variable which we will operate on
#pragma warning(suppress: 26489)
  String sTemp{std::move(sAddress)};
  Trim(sTemp);

  //divide the substring into friendly names and e-mail addresses
  auto nMark{sTemp.find(_T('<'))};
  auto nMark2{sTemp.find(_T('>'))};
  if ((nMark != String::npos) && (nMark2 != String::npos) && (nMark2 > (nMark + 1)))
  {
    m_sAddress = sTemp.substr(nMark + 1, nMark2 - nMark - 1);
    Trim(m_sAddress);
#pragma warning(suppress: 26489)
    m_sFriendlyName = sTemp.substr(0, nMark);
    Trim(m_sFriendlyName);
  }
  else
  {
    nMark = sTemp.find(_T('('));
    nMark2 = sTemp.find(_T(')'));
    if ((nMark != String::npos) && (nMark2 != String::npos) && (nMark2 > (nMark + 1)))
    {
      m_sAddress = sTemp.substr(0, nMark);
      Trim(m_sAddress);
#pragma warning(suppress: 26489)
      m_sFriendlyName = sTemp.substr(nMark + 1, nMark2 - nMark - 1);
      Trim(m_sFriendlyName);
    }
    else
    {
      m_sAddress = sTemp;
      m_sFriendlyName = m_sAddress;
    }
  }
}

CMapi::CRecipient::CRecipient(_In_ const String& sFriendlyName, _In_ const String& sAddress) : m_sAddress{sAddress}, //NOLINT(modernize-pass-by-value)
                                                                                               m_sFriendlyName{sFriendlyName}
{
}

void CMapi::CRecipient::clear() noexcept
{
  m_sAddress.clear();
  m_sFriendlyName.clear();
}

void CMapi::CRecipient::Trim(_Inout_ std::string& sValue)
{
  ATL::CAtlStringA sTempValue{sValue.c_str()};
  sValue = sTempValue.Trim();
}

void CMapi::CRecipient::Trim(_Inout_ std::wstring& sValue)
{
  ATL::CAtlStringW sTempValue{sValue.c_str()};
  sValue = sTempValue.Trim();
}


CMapi::CMessage::CMessage() noexcept : m_Flags{0}
{
}

void CMapi::CMessage::clear() noexcept
{
  m_To.clear();
  m_CC.clear();
  m_BCC.clear();
  m_sSubject.clear();
  m_sBody.clear();
  m_Attachments.clear();
  m_AttachmentTitles.clear();
  m_From.clear();
  m_sDateReceived.clear();
  m_sMessageType.clear();
  m_Flags = 0;
}

void CMapi::CMessage::AddMultipleRecipients(_In_ const String& sRecipients, _Inout_ std::vector<CRecipient>& recipients)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(sRecipients.length()); //An empty string is not allowed

  //Loop through the whole string, adding recipients as they are encountered
  String sTempRecipients{sRecipients}; //Allocate a work area (don't touch parameter itself)
  bool bLeftQuotationMark{false};
  const auto length{sTempRecipients.length()};
#pragma warning(suppress: 26429 26446)
  LPTSTR pszTempRecipients{sTempRecipients.data()};
  for (String::size_type pos{0}, start=0; pos<=length; pos++)
  {
#pragma warning(suppress: 26446)
    if (bLeftQuotationMark && sTempRecipients[pos] != _T('"'))
      continue;
#pragma warning(suppress: 26446)
    if (bLeftQuotationMark && sTempRecipients[pos] == _T('"'))
    {
      bLeftQuotationMark = false;
      continue;
    }
#pragma warning(suppress: 26446)
    if (sTempRecipients[pos] == _T('"'))
    {
      bLeftQuotationMark = true;
      continue;
    }

    //Valid separators between addresses are ',' or ';'
#pragma warning(suppress: 26481)
    if ((pszTempRecipients[pos] == _T(',')) || (pszTempRecipients[pos] == _T(';')) || (pszTempRecipients[pos] == _T('\0')))
    {
#pragma warning(suppress: 26481)
      pszTempRecipients[pos] = _T('\0');
#pragma warning(suppress: 26481)
      String sTemp(&(pszTempRecipients[start]));
      CRecipient::Trim(sTemp);

      //Let the CRecipient constructor do its work
      CRecipient To{sTemp};
#pragma warning(suppress: 26489)
      if (To.m_sAddress.length())
#pragma warning(suppress: 26486 26489)
        recipients.push_back(To);

      //Move on to the next position
      start = pos + 1;
    }
  }
}

void CMapi::CMessage::AddMultipleAttachments(_In_ const String& sAttachments, _In_z_ const TCHAR* pszDelimiters)
{
  bool bMoreData{true};
  String sTempAttachments{sAttachments};
  while (bMoreData)
  {
    const auto nEndPos{sTempAttachments.find_first_of(pszDelimiters)};
    if (nEndPos != String::npos)
    {
      String sAttachment{sTempAttachments.substr(0, nEndPos)};
      CRecipient::Trim(sAttachment);
      if (sAttachment.length())
        m_Attachments.push_back(sAttachment);
#pragma warning(suppress: 26489)
      sTempAttachments = sTempAttachments.substr(nEndPos + 1);
    }
    else
    {
      CRecipient::Trim(sTempAttachments);
      if (sTempAttachments.length())
        m_Attachments.push_back(sTempAttachments);
      bMoreData = false;
    }
  }
}


CMapi::CSession::CSession() noexcept : m_hSession{0},
                                       m_nLastError{0},
                                       m_hMapi{nullptr},
                                       m_lpfnMAPILogon{nullptr},
                                       m_lpfnMAPILogoff{nullptr},
                                       m_lpfnMAPISendMail{nullptr},
                                       m_lpfnMAPISendMailW{nullptr},
                                       m_lpfnMAPIResolveName{nullptr},
                                       m_lpfnMAPIFreeBuffer{nullptr},
                                       m_lpfnMAPIReadMail{nullptr},
                                       m_lpfnMAPIFindNext{nullptr},
                                       m_lpfnMAPIDeleteMail{nullptr}
{
#pragma warning(suppress: 26447)
  Initialise();
}

CMapi::CSession::~CSession() noexcept
{
  //Logoff if logged on
  Logoff();

  //Unload the MAPI dll
  Deinitialise();
}

HMODULE CMapi::CSession::LoadLibraryFromSystem32(_In_z_ LPCTSTR lpFileName)
{
  //Get the Windows System32 directory
  ATL::CAtlString sFullPath;
  const UINT nGSD{GetSystemDirectory(sFullPath.GetBuffer(_MAX_PATH), _MAX_PATH)};
  sFullPath.ReleaseBuffer();
  if (nGSD == 0)
#pragma warning(suppress: 26487)
    return nullptr;

    //Setup the full path and delegate to LoadLibrary
  sFullPath += _T("\\");
  sFullPath += lpFileName;
  return LoadLibrary(sFullPath);
}

void CMapi::CSession::Initialise()
{
  //Load up the MAPI dll and get the function pointers we are interested in
  m_hMapi = LoadLibraryFromSystem32(_T("MAPI32.DLL"));
  if (m_hMapi != nullptr)
  {
#pragma warning(suppress: 26490)
    m_lpfnMAPILogon = reinterpret_cast<LPMAPILOGON>(GetProcAddress(m_hMapi, "MAPILogon"));
#pragma warning(suppress: 26490)
    m_lpfnMAPILogoff = reinterpret_cast<LPMAPILOGOFF>(GetProcAddress(m_hMapi, "MAPILogoff"));
#pragma warning(suppress: 26490)
    m_lpfnMAPISendMail = reinterpret_cast<LPMAPISENDMAIL>(GetProcAddress(m_hMapi, "MAPISendMail"));
#pragma warning(suppress: 26490)
    m_lpfnMAPISendMailW = reinterpret_cast<LPMAPISENDMAILW>(GetProcAddress(m_hMapi, "MAPISendMailW"));
#pragma warning(suppress: 26490)
    m_lpfnMAPIResolveName = reinterpret_cast<LPMAPIRESOLVENAME>(GetProcAddress(m_hMapi, "MAPIResolveName"));
#pragma warning(suppress: 26490)
    m_lpfnMAPIFreeBuffer = reinterpret_cast<LPMAPIFREEBUFFER>(GetProcAddress(m_hMapi, "MAPIFreeBuffer"));
#pragma warning(suppress: 26490)
    m_lpfnMAPIReadMail = reinterpret_cast<LPMAPIREADMAIL>(GetProcAddress(m_hMapi, "MAPIReadMail"));
#pragma warning(suppress: 26490)
    m_lpfnMAPIFindNext = reinterpret_cast<LPMAPIFINDNEXT>(GetProcAddress(m_hMapi, "MAPIFindNext"));
#pragma warning(suppress: 26490)
    m_lpfnMAPIDeleteMail = reinterpret_cast<LPMAPIDELETEMAIL>(GetProcAddress(m_hMapi, "MAPIDeleteMail"));

    //If any of the functions are not available (bar m_lpfnMAPISendMailW) then fail the load
    if ((m_lpfnMAPILogon == nullptr) || (m_lpfnMAPILogoff == nullptr) ||
        (m_lpfnMAPISendMail == nullptr) || (m_lpfnMAPIResolveName == nullptr) || (m_lpfnMAPIFreeBuffer == nullptr) ||
        (m_lpfnMAPIReadMail == nullptr) || (m_lpfnMAPIFindNext == nullptr) || (m_lpfnMAPIDeleteMail == nullptr))
      Deinitialise();
  }
}

void CMapi::CSession::Deinitialise() noexcept
{
  if (m_hMapi != nullptr)
  {
    //Unload the MAPI dll and reset the function pointers to null
    FreeLibrary(m_hMapi);
    m_hMapi = nullptr;
    m_lpfnMAPILogon = nullptr;
    m_lpfnMAPILogoff = nullptr;
    m_lpfnMAPISendMail = nullptr;
    m_lpfnMAPISendMailW = nullptr;
    m_lpfnMAPIResolveName = nullptr;
    m_lpfnMAPIFreeBuffer = nullptr;
    m_lpfnMAPIReadMail = nullptr;
    m_lpfnMAPIFindNext = nullptr;
    m_lpfnMAPIDeleteMail = nullptr;
  }
}

#pragma warning(suppress: 26461)
_Success_(return != false) bool CMapi::CSession::Logon(_In_ const std::string& sProfileName, _In_ const std::string& sPassword, _In_opt_ HWND hParentWnd, _In_ bool bForceDownload)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPILogon != nullptr); //Function pointer must be valid

  //Initialise the function return value
  bool bSuccess{false};

  //Just in case we are already logged in
  Logoff();

  //Get the length of the profile name
  const auto nProfileLength{sProfileName.length()};

  //Setup the flags & UIParam parameters used in the MapiLogon call
  FLAGS flags{0};
  ULONG_PTR nUIParam{0};
  if (nProfileLength == 0)
  {
    //No profile name given, then we must interactively request a profile name
    if (hParentWnd != nullptr)
    {
#pragma warning(suppress: 26490)
      nUIParam = reinterpret_cast<ULONG_PTR>(hParentWnd);
      flags |= MAPI_LOGON_UI;
    }
  }
  if (bForceDownload)
    flags |= MAPI_FORCE_DOWNLOAD;

  //First try to acquire a new MAPI session using the supplied settings using the MAPILogon function
  std::string sTempProfileName{sProfileName};
  std::string sTempPassword{sPassword};
  m_nLastError = m_lpfnMAPILogon(nUIParam, sTempProfileName.data(), sTempPassword.data(), flags | MAPI_NEW_SESSION, 0, &m_hSession);
  if ((m_nLastError != SUCCESS_SUCCESS) && (m_nLastError != MAPI_E_USER_ABORT))
  {
    const FLAGS fSharedFlags{bForceDownload ? MAPI_FORCE_DOWNLOAD : 0u};

    //Failed to create a create mapi session, try to acquire a shared mapi session
    m_nLastError = m_lpfnMAPILogon(nUIParam, nullptr, nullptr, fSharedFlags, 0, &m_hSession);
    if (m_nLastError == SUCCESS_SUCCESS)
      bSuccess = true;
  }
  else if (m_nLastError == SUCCESS_SUCCESS)
    bSuccess = true;

  return bSuccess;
}

_Success_(return != false) bool CMapi::CSession::SharedLogon(_In_ bool bForceDownload) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPILogon != nullptr); //Function pointer must be valid

  //Initialise the function return value
  bool bSuccess{false};

  //Just in case we are already logged in
  Logoff();

  //Try to acquire a shared mapi session
  const FLAGS flags{bForceDownload ? MAPI_FORCE_DOWNLOAD : 0u};
  m_nLastError = m_lpfnMAPILogon(0, nullptr, nullptr, flags, 0, &m_hSession);
  if (m_nLastError == SUCCESS_SUCCESS)
    bSuccess = true;

  return bSuccess;
}

_Success_(return != false) bool CMapi::CSession::LoggedOn() const noexcept
{
  return (m_hSession != 0);
}

_Success_(return != false) bool CMapi::CSession::MapiInstalled() const noexcept
{
  return (m_hMapi != nullptr);
}

_Success_(return != false) bool CMapi::CSession::UnicodeSupportInstalled() const noexcept
{
  return (m_lpfnMAPISendMailW != nullptr);
}

_Success_(return != false) bool CMapi::CSession::Logoff() noexcept
{
  //Initialise the function return value
  bool bSuccess{false};

  if (m_hSession)
  {
#pragma warning(suppress: 26477)
    ATLASSERT(m_lpfnMAPILogoff != nullptr); //Function pointer must be valid

    //Call the low level function
    m_nLastError = m_lpfnMAPILogoff(m_hSession, 0, 0, 0);
    if (m_nLastError != SUCCESS_SUCCESS)
      bSuccess = false;
    else
      bSuccess = true;
    m_hSession = NULL;
  }

  return bSuccess;
}

_Success_(return != false) bool CMapi::CSession::Resolve(_In_ LPCSTR pszName, _Out_ lpMapiRecipDesc* lppRecip)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIResolveName != nullptr); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_hSession != NULL); //MAPI session handle must be valid

  //Call the low level function
  std::string sTempName{pszName};
#pragma warning(suppress: 26446)
  m_nLastError = m_lpfnMAPIResolveName(m_hSession, 0, sTempName.data(), 0, 0, lppRecip);

  return (m_nLastError == SUCCESS_SUCCESS);
}

void CMapi::CSession::SetupRecipient(_Inout_ CRecipient& recipient, _Inout_ CMapiMessageCleanup& mapiMessage, _In_ int nIndex, _In_ bool bResolve, _In_ ULONG nRecipientClass)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSUME(static_cast<ULONG>(nIndex) < mapiMessage.nRecipCount);

#pragma warning(suppress: 26481)
  MapiRecipDescT& recip{mapiMessage.lpRecips[nIndex]};
  memset(&recip, 0, sizeof(recip));
  recip.ulRecipClass = nRecipientClass;
  if (bResolve)
  {
    //Try to resolve the name
    lpMapiRecipDesc lpTempRecip{nullptr};
#pragma warning(suppress: 26486 26489)
    if (Resolve(ATL::CT2A{recipient.m_sFriendlyName.c_str()}, &lpTempRecip))
    {
      //Resolve worked, put the resolved name back into the sName
      recipient.m_sFriendlyName = ATL::CA2T{lpTempRecip->lpszName};
    #pragma warning(suppress: 26489)
      recipient.m_sAddress = ATL::CA2T{lpTempRecip->lpszAddress};

      //Don't forget to free up the memory MAPI allocated for us
      m_lpfnMAPIFreeBuffer(lpTempRecip);
    }
  }
  recip.lpszName = recipient.m_sFriendlyName.data();
  recip.lpszAddress = recipient.m_sAddress.data();
}

void CMapi::CSession::SetupRecipients(_Inout_ CMapiMessageCleanup& mapiMessage, _Inout_ CMessage& message, _In_ bool bResolve)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIFreeBuffer != nullptr);

  //Allocate the recipients array
  mapiMessage.m_Recipients.clear();
  if (mapiMessage.nRecipCount)
  {
    mapiMessage.m_Recipients.resize(mapiMessage.nRecipCount);
    mapiMessage.lpRecips = mapiMessage.m_Recipients.data();

    //Setup the "To" recipients
    int nRecipIndex{0};
    for (auto& recipient : message.m_To)
    {
      SetupRecipient(recipient, mapiMessage, nRecipIndex, bResolve, MAPI_TO);
      ++nRecipIndex;
    }

    //Setup the "CC" recipients
    for (auto& recipient : message.m_CC)
    {
      SetupRecipient(recipient, mapiMessage, nRecipIndex, bResolve, MAPI_CC);
      ++nRecipIndex;
    }

    //Setup the "BCC" recipients
    for (auto& recipient : message.m_BCC)
    {
      SetupRecipient(recipient, mapiMessage, nRecipIndex, bResolve, MAPI_BCC);
      ++nRecipIndex;
    }
  }
}

#pragma warning(suppress: 26440)
void CMapi::CSession::SetupAttachment(_In_ String& sAttachment, _Inout_ MapiMessageT& mapiMessage, _Inout_ CMapi::CMessage& message, _In_ int nIndex)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSUME(static_cast<ULONG>(nIndex) < mapiMessage.nFileCount);

#pragma warning(suppress: 26481)
  memset(&(mapiMessage.lpFiles[nIndex]), 0, sizeof(mapiMessage.lpFiles[nIndex]));
#pragma warning(suppress: 26472 26481)
  mapiMessage.lpFiles[nIndex].nPosition = static_cast<ULONG>(-1);
#pragma warning(suppress: 26481 26446)
  mapiMessage.lpFiles[nIndex].lpszPathName = sAttachment.data();
#pragma warning(suppress: 26446)
  String& sAttachmentTitle = message.m_AttachmentTitles[nIndex];
  if (sAttachmentTitle.length())
#pragma warning(suppress: 26446 26481)
    mapiMessage.lpFiles[nIndex].lpszFileName = sAttachmentTitle.data();
}

void CMapi::CSession::SetupAttachments(_Inout_ CMapiMessageCleanup& mapiMessage, _Inout_ CMapi::CMessage& message, _In_ bool bSendAsHTML, _In_z_ LPCTSTR pszTempHTMLFile)
{
  //Make the attachment titles array the same size as the attachment array if they are not the same size
  auto nAttachmentSize{message.m_Attachments.size()};
  const auto nTitleSize{message.m_AttachmentTitles.size()};
  if (nAttachmentSize && (nTitleSize != nAttachmentSize))
    message.m_AttachmentTitles.resize(nAttachmentSize);

  //If the message is to be sent in HTML, add an extra temp file
  if (bSendAsHTML)
  {
    message.m_Attachments.insert(message.m_Attachments.begin(), pszTempHTMLFile);
    message.m_AttachmentTitles.insert(message.m_AttachmentTitles.begin(), String{_T("default.htm")});
    ++nAttachmentSize;
  }

  mapiMessage.m_Files.clear();
  if (nAttachmentSize)
  {
#pragma warning(suppress: 26472)
    mapiMessage.nFileCount = static_cast<ULONG>(nAttachmentSize);
    mapiMessage.m_Files.resize(nAttachmentSize);
    mapiMessage.lpFiles = mapiMessage.m_Files.data();
    int nIndex{0};
    for (auto& sAttachment : message.m_Attachments)
#pragma warning(suppress: 26486 26489)
      SetupAttachment(sAttachment, mapiMessage, message, nIndex++);
  }
}

#pragma warning(suppress: 26461)
_Success_(return != false) bool CMapi::CSession::Send(_Inout_ CMapi::CMessage& message, _In_ bool bResolve, _In_ bool bInteractive, _In_opt_ HWND hParentWnd, _In_ bool bSendAsHTML)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT((m_lpfnMAPISendMail != nullptr) || (m_lpfnMAPISendMailW != nullptr)); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIFreeBuffer != nullptr); //Function pointer must be valid

#ifdef _UNICODE
  //Check MapiSendMailW is available
  if (m_lpfnMAPISendMailW == nullptr)
  {
    m_nLastError = MAPI_E_UNICODE_NOT_SUPPORTED;
    return false;
  }
#endif //#ifdef _UNICODE

  //Initialise the function return value
  bool bSuccess{false};

#pragma warning(suppress: 26446)
  LPTSTR pszBody{message.m_sBody.data()};

  //If the message is to be sent in HTML, setup a file, copy the body text in it and add it to the attachments
  //We create the temp file, copy the body text in it and add it to the MapiMessage as the first file attachment,
  //then we add the 'original' attachments
  ATL::CAtlString sTempFileName;
  if (bSendAsHTML)
  {
    //Get the temp path
    ATL::CAtlString sFullPath;
    const DWORD dwGTP{GetTempPath(_MAX_PATH, sFullPath.GetBuffer(_MAX_PATH))};
    sFullPath.ReleaseBuffer();
    if (dwGTP == 0)
    {
      m_nLastError = MAPI_E_ATTACHMENT_WRITE_FAILURE;
      return false;
    }

    //Now form a unique filename using GetTempFile
    const UINT nGTF{GetTempFileName(sFullPath, _T("cma"), 0, sTempFileName.GetBuffer(_MAX_PATH))};
    sTempFileName.ReleaseBuffer();
    if (nGTF == 0)
    {
      m_nLastError = MAPI_E_ATTACHMENT_WRITE_FAILURE;
      return false;
    }

    //Now write the body text to this file
    ATL::CAtlFile file;
    HRESULT hr{file.Create(sTempFileName, GENERIC_WRITE, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN)};
    if (SUCCEEDED(hr))
    {
#ifdef _UNICODE
      ATL::CT2A szUTF8Body{pszBody, CP_UTF8};
#pragma warning(suppress: 26472)
      const auto dwBodyLength{static_cast<DWORD>(strlen(szUTF8Body))};
      DWORD dwWritten{0};
      hr = file.Write(szUTF8Body, dwBodyLength, &dwWritten);
#else
#pragma warning(suppress: 26472)
      const auto dwBodyLength{static_cast<DWORD>(strlen(pszBody))};
      DWORD dwWritten{0};
      hr = file.Write(pszBody, dwBodyLength, &dwWritten);
#endif //#ifdef _UNICODE
      file.Close();

      if (dwWritten != dwBodyLength)
        hr = ATL::AtlHresultFromWin32(ERROR_DISK_FULL);
    }
    if (FAILED(hr))
    {
      DeleteFile(sTempFileName);
      m_nLastError = MAPI_E_ATTACHMENT_WRITE_FAILURE;
      return false;
    }
  }

  //Create the MapiMessage structure to match the message parameter send into us
  CMapiMessageCleanup mapiMessage;
#pragma warning(suppress: 26446)
  mapiMessage.lpszSubject = message.m_sSubject.data();

  //Set the body to null for HTML mail; the first attached file should be the HTML <body> file in this case.
  //This is the best we can do to simulate HTML email using just Simple MAPI
  if (bSendAsHTML)
    mapiMessage.lpszNoteText = nullptr;
  else
    mapiMessage.lpszNoteText = pszBody;

#pragma warning(suppress: 26472)
  mapiMessage.nRecipCount = static_cast<ULONG>(message.m_To.size() + message.m_CC.size() + message.m_BCC.size());
#pragma warning(suppress: 26477)
  ATLASSERT(mapiMessage.nRecipCount || bInteractive); //Must have at least 1 recipient if we are not "Interactive"!
  mapiMessage.flFlags = message.m_Flags;
#pragma warning(suppress: 26486)
  SetupRecipients(mapiMessage, message, bResolve);
  SetupAttachments(mapiMessage, message, bSendAsHTML, sTempFileName);

  //Setup the parameters into the function
  FLAGS flags{0};
  ULONG_PTR uiParam{0};
  if (!LoggedOn())
    flags |= MAPI_LOGON_UI;
  if (bInteractive)
  {
    flags |= MAPI_DIALOG;
    if (hParentWnd != nullptr)
#pragma warning(suppress: 26490)
      uiParam = reinterpret_cast<ULONG_PTR>(hParentWnd);
  }

  //Call the low level function
#ifdef _UNICODE
  m_nLastError = m_lpfnMAPISendMailW(m_hSession, uiParam, &mapiMessage, flags, 0);
#else
  m_nLastError = m_lpfnMAPISendMail(m_hSession, uiParam, &mapiMessage, flags, 0);
#endif //#ifdef _UNICODE
  if (m_nLastError == SUCCESS_SUCCESS)
    bSuccess = true;

  //Delete temp file and dir if sending HTML
  if (bSendAsHTML)
    DeleteFile(sTempFileName);

  return bSuccess;
}

#pragma warning(suppress: 26461)
_Success_(return != false) bool CMapi::CSession::Find(_Inout_ std::string& sMessageID, _In_ const std::string& sSeedMessageID, _In_ FLAGS flFlags, _In_ const std::string& sMessageType, _In_ bool bInteractive, _In_opt_ HWND hParentWnd)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIFindNext != nullptr); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_hSession != NULL); //MAPI session handle must be valid

  //Initialise the function return value
  bool bSuccess{false};

  //Form the UI param
  ULONG_PTR uiParam{0};
  if ((bInteractive) && (hParentWnd != nullptr))
#pragma warning(suppress: 26490)
    uiParam = reinterpret_cast<ULONG_PTR>(hParentWnd);

  //Form temp versions of the various strings
  std::string sTempSeedMessageID{sSeedMessageID};
  std::string sTempMessageType(sMessageType);

  //Call the low level function
  ATL::CAtlStringA sTempMessageID;
  m_nLastError = m_lpfnMAPIFindNext(m_hSession, uiParam, sTempMessageType.data(), sTempSeedMessageID.data(), flFlags, 0, sTempMessageID.GetBuffer(512));
  sTempMessageID.ReleaseBuffer();
  if (m_nLastError == SUCCESS_SUCCESS)
  {
    bSuccess = true;
    sMessageID = sTempMessageID;
  }

  return bSuccess;
}

#pragma warning(suppress: 26461)
_Success_(return != false) bool CMapi::CSession::Read(_In_ const std::string& sMessageID, _Inout_ CMapi::CMessage& message, _In_ FLAGS flFlags, _In_ bool bInteractive, _In_opt_ HWND hParentWnd)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIReadMail != nullptr); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIFreeBuffer != nullptr); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_hSession != NULL); //MAPI session handle must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(sMessageID.length()); //Message ID should have been specified

  //Initialise the function return value
  bool bSuccess{false};

  //Form the UI param
  ULONG_PTR uiParam{0};
  if (bInteractive && (hParentWnd != nullptr))
#pragma warning(suppress: 26490)
    uiParam = reinterpret_cast<ULONG_PTR>(hParentWnd);

  //Form the ascii version of the string
  std::string sTempMessageID{sMessageID};

  //Call the low level function
#pragma warning(suppress: 26429)
  MapiMessage* mapiMessage{nullptr};
#pragma warning(suppress: 26486)
  m_nLastError = m_lpfnMAPIReadMail(m_hSession, uiParam, sTempMessageID.data(), flFlags, 0, &mapiMessage);
  if (m_nLastError == SUCCESS_SUCCESS)
  {
    bSuccess = true;

    message.clear();
#pragma warning(suppress: 26477)
    ATLASSERT(mapiMessage->lpOriginator != nullptr);
    if (mapiMessage->lpOriginator->lpszName != nullptr)
#pragma warning(suppress: 26486)
      message.m_From.m_sFriendlyName = ATL::CA2T{mapiMessage->lpOriginator->lpszName};
    if (mapiMessage->lpOriginator->lpszAddress != nullptr)
#pragma warning(suppress: 26486)
      message.m_From.m_sAddress = ATL::CA2T{mapiMessage->lpOriginator->lpszAddress};
    if (mapiMessage->lpszSubject != nullptr)
      message.m_sSubject = ATL::CA2T{mapiMessage->lpszSubject};
    if (mapiMessage->lpszNoteText != nullptr)
      message.m_sBody = ATL::CA2T{mapiMessage->lpszNoteText};
    if (mapiMessage->lpszDateReceived != nullptr)
      message.m_sDateReceived = ATL::CA2T{mapiMessage->lpszDateReceived};
    message.m_Flags = mapiMessage->flFlags;
    if (mapiMessage->lpszMessageType != nullptr)
      message.m_sMessageType = ATL::CA2T{mapiMessage->lpszMessageType};

    //Get the recipients
    CRecipient MapiRecipient;
    for (ULONG i{0}; i<mapiMessage->nRecipCount; i++)
    {
#pragma warning(suppress: 26481)
      const MapiRecipDesc& recip{mapiMessage->lpRecips[i]};
#pragma warning(suppress: 26486 26489)
      MapiRecipient.m_sFriendlyName = ATL::CA2T{recip.lpszName};
#pragma warning(suppress: 26486 26489)
      MapiRecipient.m_sAddress = ATL::CA2T{recip.lpszAddress};
      switch (recip.ulRecipClass)
      {
        case MAPI_TO:
        {
          message.m_To.push_back(MapiRecipient);
          break;
        }
        case MAPI_CC:
        {
          message.m_CC.push_back(MapiRecipient);
          break;
        }
        case MAPI_BCC:
        {
          message.m_BCC.push_back(MapiRecipient);
          break;
        }
        default:
        {
#pragma warning(suppress: 26477)
          ATLASSERT(false);
          break;
        }
      }
    }

    //Get the attachments
    message.m_Attachments.reserve(mapiMessage->nFileCount);
    message.m_AttachmentTitles.reserve(mapiMessage->nFileCount);
    for (ULONG j{0}; j<mapiMessage->nFileCount; j++) 
    {
#pragma warning(suppress: 26481)
      const MapiFileDesc& file{mapiMessage->lpFiles[j]};
      if (file.lpszPathName != nullptr)
#pragma warning(suppress: 26486)
        message.m_Attachments.emplace_back(ATL::CA2T{file.lpszPathName});
      else
        message.m_Attachments.emplace_back();
      if (file.lpszFileName != nullptr)
#pragma warning(suppress: 26486)
        message.m_AttachmentTitles.emplace_back(ATL::CA2T{file.lpszFileName});
      else
        message.m_AttachmentTitles.emplace_back();
    }

    //Free up the mapimessage structure returned by MAPIReadMail
    m_lpfnMAPIFreeBuffer(mapiMessage);
  }

  return bSuccess;
}

#pragma warning(suppress: 26461)
_Success_(return != false) bool CMapi::CSession::Delete(_In_ const std::string& sMessageID, _In_ bool bInteractive, _In_opt_ HWND hParentWnd)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(MapiInstalled()); //MAPI must be installed
#pragma warning(suppress: 26477)
  ATLASSERT(m_lpfnMAPIDeleteMail != nullptr); //Function pointer must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(m_hSession != NULL); //MAPI session handle must be valid
#pragma warning(suppress: 26477)
  ATLASSERT(sMessageID.length() > 0); //Message ID should have been specified

  //Initialise the function return value
  bool bSuccess{false};

  //Form the UI param
  ULONG_PTR uiParam{0};
  if (bInteractive && (hParentWnd != nullptr))
#pragma warning(suppress: 26490)
    uiParam = reinterpret_cast<ULONG_PTR>(hParentWnd);

  //Form the ascii version of the string
  std::string sTempMessageID{sMessageID};

  //Call the function
  m_nLastError = m_lpfnMAPIDeleteMail(m_hSession, uiParam, sTempMessageID.data(), 0, 0);
  if (m_nLastError == SUCCESS_SUCCESS)
    bSuccess = true;

  return bSuccess;
}

ULONG CMapi::CSession::GetLastError() const noexcept
{
  return m_nLastError;
}
