/*
Module : CMapi.h
Purpose: Defines the interface for an C++ wrapper class for sending an email using simple MAPI
Created: PJN / 11-05-1999

Copyright (c) 1999 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////////// Macros / Defines ////////////////////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error CMapi requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __CMAPI_H__
#define __CMAPI_H__

#ifndef MAPI_H
#include <MAPI.h>
#pragma message("To avoid this message, please put MAPI.h in your pre compiled header (normally stdafx.h)")
#endif //#ifndef MAPI_H

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef __ATLSTR_H__
#include <atlstr.h>
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#endif //#ifndef __ATLSTR_H__

#ifndef CMAPI_EXT_CLASS
#define CMAPI_EXT_CLASS
#endif //#ifndef CMAPI_EXT_CLASS


////////////////////////// Classes ////////////////////////////////////////////

namespace CMapi
{

//Typedefs
#ifdef _UNICODE
  using String = std::wstring;
  using MapiMessageT = MapiMessageW;
  using MapiRecipDescT = MapiRecipDescW;
  using MapiFileDescT = MapiFileDescW;
#else
  using String = std::string;
  using MapiMessageT = MapiMessage;
  using MapiRecipDescT = MapiRecipDesc;
  using MapiFileDescT = MapiFileDesc;
#endif //#ifdef _UNICODE


//Class which encapsulates a MAPI recipient
class CMAPI_EXT_CLASS CRecipient
{
public:
//Constructors / Destructors
  CRecipient() = default;
  CRecipient(_In_ String sAddress);
  CRecipient(_In_ const String& sFriendlyName, _In_ const String& sAddress);
  CRecipient(const CRecipient&) = default;
  CRecipient(CRecipient&&) = default;
  ~CRecipient() = default;

//Methods
  void clear() noexcept;
  CRecipient& operator=(const CRecipient&) = default;
  CRecipient& operator=(CRecipient&&) = default;

//Member variables
  String m_sAddress; //The actual provider specific email address. For example if using the SMTP 
                     //transport provider such as in Outlook express, then this value would be
                     //"SMTP:pjna@naughter.com"
  String m_sFriendlyName; //The friendly name of the recipient e.g. "PJ Naughter"

protected:
//Methods
  static void Trim(_Inout_ std::string& sValue);
  static void Trim(_Inout_ std::wstring& sValue);

  friend class CMessage;
};


//Class which encapsulates a MAPI mail message
class CMAPI_EXT_CLASS CMessage
{
public:
//Constructors / Destructors
  CMessage() noexcept;
  CMessage(const CMessage&) = default;
  CMessage(CMessage&&) = default;
  ~CMessage() = default;

//Enums
  enum class RECIPIENT_TYPE
  {
    TO,
    CC,
    BCC
  };

//Methods
  void clear() noexcept;
  static void AddMultipleRecipients(_In_ const String& sRecipients, _Inout_ std::vector<CRecipient>& recipients);
  void AddMultipleAttachments(_In_ const String& sAttachments, _In_z_ const TCHAR* pszDelimiters = _T(",;"));
  CMessage& operator=(const CMessage&) = default;
  CMessage& operator=(CMessage&&) = default;

//Member variables
  std::vector<CRecipient> m_To; //The To: recipients
  std::vector<CRecipient> m_CC; //The CC: recipients
  std::vector<CRecipient> m_BCC; //The BCC recipients
  String m_sSubject; //The subject of the message
  String m_sBody; //The body of the message
  std::vector<String> m_Attachments; //Files to attach to the email
  std::vector<String> m_AttachmentTitles; //Titles to use for the email file attachments
  CRecipient m_From; //The sender
  String m_sDateReceived; //Date received of the message
  String m_sMessageType; //The MAPI message type
  FLAGS m_Flags; //The flags for this message (Corresponds to MapiMessage::flFlags)

protected:
  friend class CSession;
};

//Class which manages auto cleanup of a "MapiMessageT" struct as used by CMapiSession
class CMapiMessageCleanup : public MapiMessageT
{
public:
//Constructors / Destructors
  CMapiMessageCleanup() noexcept
  {
    memset(this, 0, sizeof(MapiMessageT));
  }
  CMapiMessageCleanup(const CMapiMessageCleanup&) = delete;
  CMapiMessageCleanup(CMapiMessageCleanup&&) = delete;
  ~CMapiMessageCleanup() = default;

//Methods
  CMapiMessageCleanup& operator=(const CMapiMessageCleanup&) = delete;
  CMapiMessageCleanup& operator=(CMapiMessageCleanup&&) = delete;

//Member variables
  std::vector<MapiRecipDescT> m_Recipients;
  std::vector<MapiFileDescT> m_Files;
};

//The class which encapsulates the MAPI connection
class CMAPI_EXT_CLASS CSession
{
public:
//Constructors / Destructors
  CSession() noexcept;
  CSession(const CSession&) = delete;
  CSession(CSession&&) = delete;
  virtual ~CSession() noexcept;

//Methods
  CSession& operator=(const CSession&) = delete;
  CSession& operator=(CSession&&) = delete;
  _Success_(return != false) bool Logon(_In_ const std::string& pszProfileName, _In_ const std::string& sPassword = std::string(""), _In_opt_ HWND hParentWnd = nullptr, _In_ bool bForceDownload = false);
  _Success_(return != false) bool SharedLogon(_In_ bool bForceDownload = false) noexcept;
  [[nodiscard]] _Success_(return != false) bool LoggedOn() const noexcept;
  _Success_(return != false) bool Logoff() noexcept;
  _Success_(return != false) bool Send(_Inout_ CMessage& message, _In_ bool bResolve = false, _In_ bool bInteractive = false, _In_opt_ HWND hParentWnd = nullptr, _In_ bool bSendAsHTML = false);
  _Success_(return != false) bool Find(_Inout_ std::string& sMessageID, _In_ const std::string& sSeedMessageID = std::string(""), _In_ FLAGS flFlags = MAPI_LONG_MSGID, _In_ const std::string& sMessageType = std::string(""), _In_ bool bInteractive = false, _In_opt_ HWND hParentWnd = nullptr);
  _Success_(return != false) bool Read(_In_ const std::string& sMessageID, _Inout_ CMessage& message, _In_ FLAGS flFlags = MAPI_SUPPRESS_ATTACH, _In_ bool bInteractive = false, _In_opt_ HWND hParentWnd = nullptr);
  _Success_(return != false) bool Delete(_In_ const std::string& sMessageID, _In_ bool bInteractive = false, _In_opt_ HWND hParentWnd = nullptr);
  [[nodiscard]] _Success_(return != false) bool MapiInstalled() const noexcept;
  [[nodiscard]] _Success_(return != false) bool UnicodeSupportInstalled() const noexcept;
  [[nodiscard]] ULONG GetLastError() const noexcept;

protected:
//Methods
  void Initialise();
  void Deinitialise() noexcept;
  [[nodiscard]] HMODULE LoadLibraryFromSystem32(_In_z_ LPCTSTR lpFileName);
  _Success_(return != false) bool Resolve(_In_ LPCSTR pszName, _Out_ lpMapiRecipDesc* lppRecip);
  virtual void SetupRecipients(_Inout_ CMapiMessageCleanup& mapiMessage, _Inout_ CMessage& message, _In_ bool bResolve);
  virtual void SetupRecipient(_Inout_ CRecipient& recipient, _Inout_ CMapiMessageCleanup& mapiMessage, _In_ int nIndex, _In_ bool bResolve, _In_ ULONG nRecipientClass);
  virtual void SetupAttachments(_Inout_ CMapiMessageCleanup& mapiMessage, _Inout_ CMessage& message, _In_ bool bSendAsHTML, _In_z_ LPCTSTR pszTempHTMLFile);
  virtual void SetupAttachment(_In_ String& sAttachment, _Inout_ MapiMessageT& mapiMessage, _Inout_ CMessage& message, _In_ int nIndex);

//Data
  LHANDLE m_hSession; //Mapi Session handle
  ULONG m_nLastError; //Last Mapi error value
  HINSTANCE m_hMapi; //Instance handle of the MAPI dll
  LPMAPILOGON m_lpfnMAPILogon; //MAPILogon function pointer
  LPMAPILOGOFF m_lpfnMAPILogoff; //MAPILogoff function pointer
  LPMAPISENDMAIL m_lpfnMAPISendMail; //MAPISendMail function pointer
  LPMAPISENDMAILW m_lpfnMAPISendMailW; //MapiSendMailW function pointer
  LPMAPIRESOLVENAME m_lpfnMAPIResolveName; //MAPIResolveName function pointer
  LPMAPIFREEBUFFER m_lpfnMAPIFreeBuffer; //MAPIFreeBuffer function pointer
  LPMAPIREADMAIL m_lpfnMAPIReadMail; //MAPIReadMail function pointer
  LPMAPIFINDNEXT m_lpfnMAPIFindNext; //MAPIFindNext function pointer
  LPMAPIDELETEMAIL m_lpfnMAPIDeleteMail; //MAPIDeleteMail function pointer
};


}; //namespace CMapi


#endif //#ifndef __CMAPI_H__
